#include <string.h>
#include "printf.h"
#include "msioLL.h"
#include "msioHL.h"
#include "msPro.h"
#include "timer.h"
#include "util.h"

//MSPRO regs
#define MS_PRO_REG_INT				0x01	//RO
#define MS_PRO_REG_STATUS			0x02	//RO
#define MS_PRO_REG_TYPE				0x04	//RO
#define MS_PRO_REG_CATEGORY			0x06	//RO
#define MS_PRO_REG_CLASS			0x07	//RO
#define MS_PRO_REG_SYSPARAM			0x10	//WO
#define MS_PRO_REG_DAT_CNT_1		0x11	//WO		num 512-byte "units" to xfer. BE. if zero, read/write will not end till MS_PRO_CMD_STOP
#define MS_PRO_REG_DAT_CNT_0		0x12	//WO
#define MS_PRO_REG_DAT_ADR_3		0x13	//WO		address (in units?). BE
#define MS_PRO_REG_DAT_ADR_2		0x14	//WO
#define MS_PRO_REG_DAT_ADR_1		0x15	//WO
#define MS_PRO_REG_DAT_ADR_0		0x16	//WO
#define MS_PRO_REG_TPC_PARAM		0x17	//WO		short xfer size in lower 2 bits: {0->32, 1->64, 2->128, 3->256}
#define MS_PRO_REG_CMD_PARAM		0x18	//WO

//MSPRO commands
#define MS_PRO_CMD_FORMAT			0x10
#define MS_PRO_CMD_SLEEP			0x11
#define MS_PRO_CMD_READ				0x20
#define MS_PRO_CMD_WRITE			0x21
#define MS_PRO_CMD_ATTRIB_READ		0x24
#define MS_PRO_CMD_STOP				0x25
#define MS_PRO_CMD_ERASE			0x26


bool msproIsMspro(const uint8_t *firstEightRegs)
{
	return firstEightRegs[MS_PRO_REG_TYPE] == 0x01;
}

static bool msproPrvCmdEx(uint_fast8_t cmd, uint_fast16_t count, uint32_t addr)	//waits for 0x20 INT reg, which may not be proper for MS_PRO_CMD_STOP
{	
	uint8_t rx;
	
	if (!msioHlCmdEx(cmd, addr, count)) {
		pr("cmd broken\n");
		return false;
	}
	
	if (!msioHlCmdEndWaitInt(TIMER_TICKS_PER_SEC * 15 / 10, &rx)) {
		pr("comand timed out\n");
		return false;
	}
	
	pr("INT: %02x\n", rx);
	
	return rx == 0x20;
}

static bool msPrvReadMsProAttrPage(uint_fast8_t pgIdx, uint8_t *buf)		//maybe MsPro only, buffer should be 514 bytes
{
	return msproPrvCmdEx(MS_PRO_CMD_ATTRIB_READ, 1, pgIdx) && msioLlTpcR(MS_RD_LDATA, buf, 512);
}

static bool msPrvReadAttr(uint32_t ofst, uint32_t len, uint8_t *dst)
{
	uint8_t buf[512];
	uint32_t now;
	
	while (len) {
		now = 512 - (ofst % 512);
		if (now > len)
			now = len;
		
		if (!msPrvReadMsProAttrPage(ofst / 512, buf))
			return false;
		memcpy(dst, buf + ofst % 512, now);
		dst += now;
		ofst += now;
		len -= now;
	}
	
	return true;
}

static void msPrvShowMsproAttr(uint_fast8_t id, const uint8_t *arr, uint32_t len)
{
	uint_fast8_t i;
	
	switch (id) {
		case 0x10:
			if (len != 96) {
				pr("invalid length\n");
				break;
			}
			pr("System Information block\n");
			pr(" Memory stick class:    %u\n", arr[0]);
			pr(" Block size:            %u\n", readBE16(arr + 2));
			pr(" Total blocks:          %u\n", readBE16(arr + 4));
			pr(" User area blocks:      %u\n", readBE16(arr + 6));
			pr(" Page Size:             %u\n", readBE16(arr + 8));
			pr(" Assembly date:         %04u.%02u.%02u %02u:%02u:%02u\n", readBE16(arr + 13), arr[15], arr[16], arr[17], arr[18], arr[19]);
			pr(" Serial number:         %u\n", readBE32(arr + 20));
			pr(" Assembly maker code:   %02xh\n", arr[24]);
			pr(" Assembly model code:   %02xh %02xh %02xh\n", arr[25], arr[26], arr[27]);
			pr(" Memory maker code:     %02xh %02xh\n", arr[28], arr[29]);
			pr(" Memory model code:     %02xh %02xh\n", arr[30], arr[31]);
			pr(" Vcc:                   %u\n", arr[36]);
			pr(" Vpp:                   %u\n", arr[37]);
			pr(" Controller number:     %02xh %02xh\n", arr[38], arr[39]);
			pr(" Controller function:   %02xh %02xh\n", arr[40], arr[41]);
			pr(" Start sector:          %u\n", readBE16(arr + 42));
			pr(" Unit size:             %u\n", readBE16(arr + 44));
			pr(" Memory stick subclass: %02xh\n", arr[46]);
			pr(" Interface type:        %02xh\n", arr[51]);
			pr(" Controller code:       %02xh %02xh\n", arr[52], arr[53]);
			pr(" Format type:           %02xh\n", arr[54]);
			pr(" Device type:           %02xh\n", arr[56]);
			pr(" Memory Stick PRO ID:  ");
			for (i = 0; i < 16; i++)
				pr(" %02x", arr[64 + i]);
			pr("\n");
			pr(" TOTAL SIZE:            %u sectors (%llu bytes)\n",
				(uint32_t)readBE16(arr + 2) * readBE16(arr + 6),
				((uint64_t)((uint32_t)readBE16(arr + 2) * readBE16(arr + 6))) * 512);
			break;
		
		case 0x11:
		case 0x12:
			pr("vendor info %u\n", id - 0x10);
			break;
		
		case 0x15:
			pr(" MODEL: '%s'\n", arr);
			break;
		
		case 0x20:
			pr("MBR values\n");
			break;
		
		case 0x21:
		case 0x22:
			pr("PBR values FAT%u\n", 16 << (id - 0x21));
			break;
		
		case 0x25:
		case 0x26:
			pr("Special file values %u\n", id - 0x24);
			break;
		
		case 0x30:
			pr("Identify Device Information\n");
			pr(" Number of cylinders:   %u\n", readBE16(arr + 0));
			pr(" Number of heads:       %u\n", readBE16(arr + 2));
			pr(" Unformatted B/track:   %u\n", readBE16(arr + 4));
			pr(" Unformatted B/sector:  %u\n", readBE16(arr + 6));
			pr(" sectors/track:         %u\n", readBE16(arr + 8));
			break;
	
		default:
			pr("unknown attribute 0x%02x\n", id);
			break;
	}	
}

void msproDemo(void)
{
	uint_fast8_t numAttrs, i;
	uint8_t vals[512];
	
	if (!msioHlSetRegAddrs(0x00, 0x08, 0x10, 0x10)) {
		pr("failed to set reg addrs\n");
		return;
	}
	
	//wait for init
	do {
		(void)msioHlRegsRead(vals, 8);
	} while (!(vals[1] & 0x80));
	
	pr("REGS:");
	showBuf(vals, 8);
	pr("card inited\n");
	
	if (!msPrvReadMsProAttrPage(0, vals)) {
		pr("failed to read attrs page\n");
		return;
	}
	showBuf(vals, 512);
	
	//parse attribs
	if (vals[0] != 0xa5 || vals[1] != 0xc3 || vals[2] != 0x01 || (numAttrs = vals[4]) > 12) {
		pr("attrib page unparseable\n");
		return;
	}
	for (i = 0; i < numAttrs; i++) {
		uint32_t ofst = readBE32(vals + 16 + 12 * i + 0);
		uint32_t len = readBE32(vals + 16 + 12 * i + 4);
		uint_fast8_t id = vals[16 + 12 * i + 8];
		uint8_t arr[96];	//big enough?
		
		pr("getting attr %u (id %u) (0x%08x + 0x%08x)\n", i, id, ofst, len);
		
		if (ofst < 0x1a0 || ofst + len > 0x8000) {
			pr("invalid attr\n");
			return;
		}
		
		if (len > sizeof(arr)) {
			
			pr("too big\n");
			continue;
		}
		
		if (!msPrvReadAttr(ofst, len, arr)) {
			
			pr("failed to read attribute\n");
			return;
		}
		
		showBuf(arr, len);
		msPrvShowMsproAttr(id, arr, len);
	}
	
	while(1);
	
	//let's try to read some sectors?
	if (!msproPrvCmdEx(MS_PRO_CMD_READ, 2, 0)) {
		pr("failed to set up read command\n");
		return;
	}
	
	//let's read data
	if (!msioLlTpcR(MS_RD_LDATA, vals, 512)) {
		pr("failed to set up read pg 0\n");
		return;
	}
	showBuf(vals, 512);
	
	if (!msioLlTpcR(MS_RD_LDATA, vals, 512)) {
		pr("failed to set up read pg 1\n");
		return;
	}
	showBuf(vals, 512);
	
	//let's write data
	vals[5] = 0x66;
	vals[6] = 0x77;
	vals[7] = 0x88;
	
	//let's try to read some sectors?
	if (!msproPrvCmdEx(MS_PRO_CMD_WRITE, 1, 1)) {
		pr("failed to set up write command\n");
		return;
	}
	
	if (!msioLlTpcR(MS_WR_LDATA, vals, 512)) {
		pr("failed to set up wr pg 1\n");
		return;
	}
	
	pr("wrote\n");
	memset(vals, 0, sizeof(vals));
	
	pr("readback\n");
	if (!msproPrvCmdEx(MS_PRO_CMD_READ, 1, 1)) {
		pr("failed to set up read command\n");
		return;
	}
	
	if (!msioLlTpcR(MS_RD_LDATA, vals, 512)) {
		pr("failed to set up read pg 1\n");
		return;
	}
	showBuf(vals, 512);
}
