#include <string.h>
#include "msioCamera.h"
#include "printf.h"
#include "msioLL.h"
#include "msioHL.h"
#include "timer.h"
#include "util.h"


static bool msioPrvCamLlCmd(uint_fast8_t cmd0, uint_fast8_t cmd1, uint_fast8_t cmd2, uint_fast8_t paramSz, const uint8_t *params, uint_fast8_t replySz, uint8_t *replyP)
{
	uint8_t tx[13] = {replySz + 1, paramSz + 4, cmd0, cmd1, cmd2, }, rx[9];
	uint_fast8_t i;
	
	pr("  cmd.LL(%u %u %u tx: %u bytes rx %u bytes)\n", cmd0, cmd1, cmd2, paramSz, replySz);
	
	if (paramSz >= 8 || replySz >= 8) {
		pr("invalid lengths\n");
		return false;	
	}
	
	memcpy(tx + 5, params, paramSz);
	
	if (!msioHlSetRegAddrs(0x00, 0x0f, 0x25, sizeof(tx))) {
		pr("failed to set reg addrs 1\n");
		return false;
	}
	
	if (!msioHlRegsWrite(tx, sizeof(tx))) {
		pr("failed to write cmd\n");
		return false;
	}
	
	for (i = 0; i < 2; i++) {
		
		uint8_t cmdOp = i ? 0x01 : 0x10;
		
		delayUsec(200);
		
		if (!msioHlSetRegAddrs(0x33, replySz + 1, 0x32, 1)) {
			pr("failed to set reg addrs 2\n");
			return false;
		}
	
		delayUsec(100);
		
		if (!msioHlRegsWrite(&cmdOp, 1)) {
			pr("failed to write cmd op %u\n", i + 1);
			return false;
		}
		
		if (!msioHlCmdEndWaitInt(TIMER_TICKS_PER_SEC * 40 / 1000, NULL)) {
			pr("failed to wait for cmd %u\n", i + 1);
			return false;
		}
		
		delayUsec(500);
	}
	
	if (!msioHlRegsRead(rx, replySz + 1)) {
		pr("failed to read reply\n");
		return false;
	}

	if (rx[0] != replySz + 1) {
		
		pr("reply had %u bytes and we expected %u\n", rx[0], replySz + 1);
		return false;
	}
	
	memcpy(replyP, rx + 1, replySz);
	return true;
}

static bool msioPrvCamHlValRead(uint_fast8_t cmd1, uint_fast8_t cmd2, uint8_t len, uint8_t *dataOutP)
{
	pr(" rd.HL(%u %u, u bytes)\n", cmd1, cmd2, len);
	
	return msioPrvCamLlCmd(0x01, cmd1, cmd2, 1, &len, len, dataOutP);
}

static bool msioPrvCamHlValWrite(uint_fast8_t cmd1, uint_fast8_t cmd2, uint_fast8_t len, const uint8_t *data)
{
	uint8_t reply[8];
	
	pr(" wrHL(%u %u, u bytes)\n", cmd1, cmd2, len);
	
	return msioPrvCamLlCmd(0x02, cmd1, cmd2, len, data, len, reply) && !memcmp(reply, data, len);
}

enum CamHlEffect {
	CamHlEffectNone,
	CamHlEffectBlackAndWhite,
	CamHlEffectSepia,
	CamHlEffectNegative,
	CamHlEffectSolarize,
};

enum CamHlWhiteBalance {
	CamHlWhiteBalanceAuto,
	CamHlWhiteBalanceIncadescent,
	CamHlWhiteBalanceFlorescent,
	CamHlWhiteBalanceOutdoors,
};

static bool msioPrvCamHlSetWhiteBalance(enum CamHlWhiteBalance bal)
{
	static const uint8_t val[] = {
		[CamHlWhiteBalanceAuto] = 0,
		[CamHlWhiteBalanceIncadescent] = 1,
		[CamHlWhiteBalanceFlorescent] = 5,
		[CamHlWhiteBalanceOutdoors] = 7,
	};
	
	return msioPrvCamHlValWrite(2, 5, 1, &val[bal]);
}

static bool msioPrvCamHlSetEffect(enum CamHlEffect effect)
{
	static const uint8_t val[] = {
		[CamHlEffectNone] = 0x00,
		[CamHlEffectBlackAndWhite] = 0x10,
		[CamHlEffectSepia] = 0x30,
		[CamHlEffectNegative] = 0x40,
		[CamHlEffectSolarize] = 0x80,
	};
	
	return msioPrvCamHlValWrite(1, 5, 1, &val[effect]);
}

static bool msioPrvCamHlSetExposure(uint8_t val)		//1..8
{
	return msioPrvCamHlValWrite(4, 7, 1, &val);
}



static void yuv2rgb(uint8_t *rgb, uint8_t y, int8_t u, int8_t v)
{
	int32_t r, g, b;
	
	u -= 128;
	v -= 128;
	
	r = (int_fast16_t)(uint_fast16_t)y + 292 * (int_fast16_t)v / 256;
	g = (int_fast16_t)(uint_fast16_t)y - (101 * (int_fast16_t)u + 149 * (int_fast16_t)v) / 256;
	b = (int_fast16_t)(uint_fast16_t)y + 520 * (int_fast16_t)u / 256;
	
	#define clamp(x)	(x > 255 ? 255 : (x < 0 ? 0 : x))
	
	rgb[0] = clamp(r);
	rgb[1] = clamp(g);
	rgb[2] = clamp(b);
}


#define IMG_BUF_SZ		512


static void showFrame(uint_fast16_t width, uint_fast16_t height)
{
	#define NUM_ROWS_BUFFERED			2		//must be even
	
	uint8_t rows[width * 2 * NUM_ROWS_BUFFERED], buf[IMG_BUF_SZ];
	uint_fast16_t ofst, i;
	uint32_t fillProgress = 0;
		
	for (ofst = 0; ofst < width * height * 2; ofst += IMG_BUF_SZ) {
		
		uint32_t need, avail = IMG_BUF_SZ;
		uint_fast8_t retryCtr;
		uint8_t *src = buf;
		
		delayUsec(500);
	
		if (!msioLlTpcRex(MS_RD_LDATA, buf, IMG_BUF_SZ, ofst ? TIMER_TICKS_PER_SEC * 10 / 1000 : TIMER_TICKS_PER_SEC * 400 / 1000)) {
			pr("failed to read at ofst 0x%04x\n", ofst);
			break;
		}
		
		//fill the rest of the 2-line buffer)
		while (avail) {
			
			need = sizeof(rows) - fillProgress;
			if (need > avail)
				need = avail;
			
			memcpy(rows + fillProgress, src, need);
			fillProgress += need;
			avail -= need;
			src += need;
			
			if (fillProgress == sizeof(rows)) {
				
				uint_fast16_t row, col;
				uint8_t *rowTop = rows, *rowBot;
				
				//draw two rows
				fillProgress = 0;
				
				for (row = 0; row < NUM_ROWS_BUFFERED; row += 2) {
					
					rowBot = rowTop + 2 * width;
					
					for (col = 0; col < width; col += 2, rowTop += 4, rowBot += 4) {
						
						//ansi escape sequence to set background color to rgb triplet, then print a "upper block"
						//so foregound color is on top, background on bottom
						
						uint8_t rgbTop[3], rgbBot[3];
						
						yuv2rgb(rgbTop, rowTop[1], rowTop[2], rowTop[0]);
						yuv2rgb(rgbBot, rowBot[1], rowBot[2], rowBot[0]);
						
						pr("%c[38;2;%u;%u;%um%c[48;2;%u;%u;%um%c%c%c", 
								0x1b, rgbTop[0], rgbTop[1], rgbTop[2],
								0x1b, rgbBot[0], rgbBot[1], rgbBot[2],
								0xe2, 0x96, 0x80);
					
						yuv2rgb(rgbTop, rowTop[3], rowTop[2], rowTop[0]);
						yuv2rgb(rgbBot, rowBot[3], rowBot[2], rowBot[0]);
						
						pr("%c[38;2;%u;%u;%um%c[48;2;%u;%u;%um%c%c%c", 
								0x1b, rgbTop[0], rgbTop[1], rgbTop[2],
								0x1b, rgbBot[0], rgbBot[1], rgbBot[2],
								0xe2, 0x96, 0x80);
					}
					pr("%c[48;5;%um\n", 0x1b, 15);
					rowTop = rowBot;
				}
			}
		}
	}
}

static bool setCamRes(bool highRes)
{
	uint8_t reg_0x23, val_0x04, checkVal, iter, val_2_4, val8;

	if (!msioHlSingleRegRead(0x23, &reg_0x23)) {
		
		pr("failed to read reg 0x23\n");
		return false;
	}
	pr("reg 0x23: %02x\n", reg_0x23);
	
	if (!msioPrvCamHlValRead(1, 4, 1, &val_0x04)) {
		
		pr("failed to read val 0x04\n");
		return false;
	}
	pr("val 0x04: %02x\n", val_0x04);
	
	if (highRes) {
		val_0x04 &=~ 0x04;
		reg_0x23 &=~ 0x20;
	}
	else {
		val_0x04 |= 0x04;
		reg_0x23 |= 0x20;
	}
	
	if (!msioHlSingleRegWrite(0x23, reg_0x23)) {
		
		pr("failed to write reg 0x23\n");
		return false;
	}
	
	if (!msioPrvCamHlValWrite(1, 4, 1, &val_0x04)) {
		
		pr("failed to write val 0x04\n");
		return false;
	}

	for (iter = 0; iter < 16; iter++) {
		
		if (!msioPrvCamHlValRead(1, 4, 1, &checkVal)) {
			
			pr("failed to read val 0x04 again\n");
			return false;
		}
		
		if (checkVal == val_0x04)
			break;
		
		delayMsec(5);
	}
	
	if (iter == 16) {
		
		pr("failed to wait for val_0x04 to flip. wrote %02x got $%02x\n", val_0x04, checkVal);
		return false;
	}
	
	return true;
}

void msioCameraDemo(void)
{
	uint8_t buf[IMG_BUF_SZ], intRegVal;
	uint_fast16_t ofst, i;
	bool hiRes = false;
	
	delayMsec(75);
	
	//set func their way
	if (!msioHlSetRegAddrs(6, 1, 6, 1)) {
		pr("failed to set reg addrs\n");
		return;
	}
	
	buf[0] = 0x06;
	if (!msioHlRegsWrite(buf, 1)) {
		pr("failed to write regs 1\n");
		return;
	}
	
	if (!msioHlRegsRead(buf, 1)) {
		pr("failed to read regs 1\n");
		return;
	}
	
	if (buf[0] != 6) {
		
		pr("func switch fail\n");
		return;
	}
	if (!msioHlSetRegAddrs(8, 1, 8, 1)) {
		pr("failed to set reg addrs 4\n");
		return;
	}
	
	if (!msioHlRegsRead(buf, 1)) {
		pr("failed to read regs 4\n");
		return;
	}
	
	pr("Needed current: %umA\n", 10 * buf[0]);
	
	//disable function
	buf[0] = 0x00;
	if (!msioHlRegsWrite(buf, 1)) {
		pr("failed to write regs\n");
		return;
	}
	
	pr("wait a bit...\n");
	delayMsec(500);
	pr("ok, going on\n");
	
	//enable function
	buf[0] = 0xc0;
	if (!msioHlRegsWrite(buf, 1)) {
		pr("failed to write regs\n");
		return;
	}
	pr("wait a bit...\n");
	delayMsec(1500);
	pr("ok, going on\n");
	
	//init camera
	static const uint8_t settings[] = {0x00, 0x62, 0x02, 0x06};
	uint8_t reg_0x23, val_0x04, checkVal, iter, val_2_4, val8;
	
	if (!msioPrvCamHlValWrite(1, 1, sizeof(settings), settings)) {
		pr("failed to write settings\n");
		return;
	}

	delayMsec(75);
	

	if (!setCamRes(false))
		return;
	
	if (!msioPrvCamHlValRead(2, 4, 1, &val_2_4)) {
		
		pr("failed to read val 2.4\n");
		return;
	}
	val_2_4 &=~ 0x02;
	val_2_4 |= 0x04;
	val_2_4 &=~ 0x40;
	if (!msioPrvCamHlValWrite(2, 4, 1, &val_2_4)) {
		
		pr("failed to write val 2.4\n");
		return;
	}
	
	val8 = 0x01;
	if (!msioPrvCamHlValWrite(5, 3, 1, &val8)) {
		
		pr("failed to write val 5.3\n");
		return;
	}
	
	val8 = 0x73;
	if (!msioPrvCamHlValWrite(2, 29, 1, &val8)) {
		
		pr("failed to write val 2.29\n");
		return;
	}
	
	val8 = 0x93;
	if (!msioPrvCamHlValWrite(2, 30, 1, &val8)) {
		
		pr("failed to write val 2.30\n");
		return;
	}
	
	val8 = 0xC4;
	if (!msioPrvCamHlValWrite(3, 11, 1, &val8)) {
		
		pr("failed to write val 3.11\n");
		return;
	}
	
	val8 = 0x23;
	if (!msioPrvCamHlValWrite(3, 12, 1, &val8)) {
		
		pr("failed to write val 3.12\n");
		return;
	}
	
	pr("in theory, cam is configured\n");
	
	
	//setup settings
	
	msioPrvCamHlSetEffect(CamHlEffectNone);
	msioPrvCamHlSetWhiteBalance(CamHlWhiteBalanceAuto);
	msioPrvCamHlSetExposure(5);


	pr("frame data might be ready\n");
	
	while (1) {
		if (!setCamRes(hiRes))
			return;
		
		showFrame(176 * (hiRes ? 2 : 1), 144 * (hiRes ? 2 : 1));
		hiRes = !hiRes;
	}
	
	pr("done\n");
	while(1);
}


