#include "stm32g030xx.h"
#include <string.h>
#include "timebase.h"
#include "printf.h"

/*
	1 is BS
	5 is SDIO
	6 is SCLK
	
	option bits to configure:
		nBOOT_SEL = 0
		nBOOT0 = do not care (1)
		nBOOT1 = 1
	
	
	-> 0x1fff7800 to 1101 1110 1111 1111 1110 0001 1010 1010
	-> 0x1fff7800 to 0xdeffe1aa
	this is "make safefuses"
	
	disabling reset further requires:
		setting bits 27..28 to:
			0b00 - reserved
			0b01 - input reset (resets chip on input low)
			0b10 - gpio
			0b11 - i/o reset (when chip resets, it outputs low, when output is low, chip resets)
		
		default is 0b11
		
		so our state should be
	
	-> 0x1fff7800 to 1101 0110 1111 1111 1110 0001 1010 1010
	-> 0x1fff7800 to 0xd6ffe1aa
	this is "make realfuses"
	
	pin4 is nRST
	pin7 is SWDIO
	pin8 is BOOT0/SWCLK
*/

void delayUsec(uint32_t usec)
{
	uint64_t till = getTime() + (uint64_t)usec * (TICKS_PER_SECOND / 1000000);
	
	while (getTime() < till);
}

void delayMsec(uint32_t msec)
{
	uint64_t till = getTime() + (uint64_t)msec * (TICKS_PER_SECOND / 1000);
	
	while (getTime() < till);
}

void prPutchar(char chr)
{
	volatile uint32_t *addr = (volatile uint32_t*)ZWT_ADDR;
	
	while(addr[0] & 0x80000000ul) asm volatile("dsb sy");
	addr[0] = 0x80000000ul | (uint8_t)chr;
}

static void showBuf(const uint8_t *buf, uint_fast16_t len)
{
	uint_fast16_t i;
	
	for (i = 0; i < len; i++) {
		
		if (!(i & 15))
			pr("\n  %03x  ", i);
		pr(" %02x", *buf++);
	}
	pr("\n");
}

static void initHw(void)
{
	SCB->VTOR = (uint32_t)0x08000000;
	
	//enable clocks to important places
	RCC->AHBENR = RCC_AHBENR_FLASHEN;
	RCC->APBENR1 = RCC_APBENR1_DBGEN | RCC_APBENR1_PWREN;
	RCC->APBENR2 = RCC_APBENR2_SYSCFGEN;
	RCC->IOPENR = RCC_IOPENR_GPIOAEN | RCC_IOPENR_GPIOBEN | RCC_IOPENR_GPIOCEN | RCC_IOPENR_GPIODEN | RCC_IOPENR_GPIOFEN;
	RCC->CCIPR = (RCC->CCIPR &~ RCC_CCIPR_USART1SEL) | RCC_CCIPR_USART1SEL_0;	//usart1 on sysclk
	
	//setup flash
	FLASH->ACR |= FLASH_ACR_ICRST;
	FLASH->ACR = (FLASH->ACR &~ FLASH_ACR_LATENCY) | FLASH_ACR_ICEN | FLASH_ACR_PRFTEN | (2 << FLASH_ACR_LATENCY_Pos);
	
	//voltage scaling
	PWR->CR1 = PWR_CR1_VOS_0;
	while (PWR->SR2 & PWR_SR2_VOSF);
	
	//set up all AHBs and APBs with no division, use HSI (16MHz)
	RCC->CFGR = 0;
	
	//turn off the PLL
	RCC->CR &=~ RCC_CR_PLLON;
	//start PLL (input = HSI / 2 = 8MHz), VCO output 128MHz, make it output 64MHz on output "R", nothing elsewhere
	RCC->PLLCFGR = RCC_PLLCFGR_PLLSRC_HSI | RCC_PLLCFGR_PLLM_0 | ((TICKS_PER_SECOND / 4000000) << RCC_PLLCFGR_PLLN_Pos) | RCC_PLLCFGR_PLLREN | RCC_PLLCFGR_PLLR_0;
	//turn it on
	RCC->CR |= RCC_CR_PLLON;
	//wait for it
	while (!(RCC->CR & RCC_CR_PLLRDY));
	//go to it
	RCC->CFGR = RCC_CFGR_SW_1;
	
	/*
	due to multiple pads going to the same pin, we need to be careful not to misconfigure the gpio and burn out on-die transistors
	pulls also need to be carefully managed
	pin to pad map:
		1: PB7 PB8 PB9 PC14
		4: PA0 PA1 PA2 nRST
		5: PA8 PA11[9], PB0 PB1
		6: PA12[10]
		7: PA13
		8: PA14 PA15 PB5 PB6
		
		
	our use:
		1 = BS = PB7
		4 = pa0 = trigger
		5 = SDIO = PA11
		6 = SCLK = PA12
	*/
		
	GPIOA->MODER = (GPIOA->MODER &~ ((3 << (2 * 0)) | (3 << (2 * 11)) | (3 << (2 * 12)))) | (1 << (2 * 0)) | (0 << (2 * 11)) | (1 << (2 * 12));
	GPIOB->MODER = (GPIOB->MODER &~ (3 << (2 * 7))) | (1 << (2 * 7));
	GPIOA->OSPEEDR = (GPIOA->OSPEEDR &~ ((3 << (2 * 0)) | (3 << (2 * 11)) | (3 << (2 * 12)))) | (2 << (2 * 0)) | (2 << (2 * 11)) | (2 << (2 * 12));
	GPIOB->OSPEEDR = (GPIOB->OSPEEDR &~ (3 << (2 * 7))) | (2 << (2 * 7));
	
	//data has a pulldown
	GPIOA->PUPDR = (GPIOA->PUPDR &~ (3 << (2 * 11))) | (2 << (2 * 11));
	
	//trigger is low
	GPIOA->BRR = 1;
}


static void __attribute__((noinline)) msPrvDelay(void)
{
	asm volatile("nop\nnop\n");
}

//this is done on byte boundaries to allow an external logic analyzer to sync on the bytes
//this has no relation to memory stick iface
static void msPrvTrigger(void)
{
	GPIOA->BSRR = 1;
	msPrvDelay();
	GPIOA->BRR = 1;
}

static void msPrvBsToggle(void)
{
	msPrvDelay();
	if (GPIOB->ODR & (1 << 7))
		GPIOB->BRR = 1 << 7;
	else
		GPIOB->BSRR = 1 << 7;
	msPrvDelay();
}

static void msPrvSdioOutput(void)
{
	GPIOA->MODER |= (1 << (2 * 11));
}

static void msPrvSdioInput(void)
{
	GPIOA->MODER &=~ (3 << (2 * 11));
}

static void msPrvBitWr(bool hi, bool toggleBs)
{
	GPIOA->BSRR = hi ? (1 << 11) : (1 << (16 + 11));
	msPrvDelay();
	GPIOA->BSRR = 1 << 12;
	msPrvDelay();
	GPIOA->BRR = 1 << 12;
	if (toggleBs)
		msPrvBsToggle();
	msPrvDelay();
}

static void msPrvByteWr(uint_fast8_t val, bool toggleBs)
{
	uint_fast8_t i;
	
	msPrvTrigger();
	for (i = 0; i < 8; i++, val <<= 1) {
		
		msPrvBitWr((val >> 7) & 1, toggleBs && i == 6);
	}
}

static uint_fast8_t msPrvBitRd(void)
{
	uint_fast8_t ret;
	
	GPIOA->BSRR = 1 << 12;
	msPrvDelay();
	ret = (GPIOA->IDR >> 11) & 1;		//input at falling edge (rising seems to make unofficial microsd-to-mspro adapter unhappy)
	GPIOA->BRR = 1 << 12;
	msPrvDelay();
	
	return ret;
}

static uint_fast8_t msPrvByteRd(void)
{
	uint_fast8_t i, ret = 0;
	
	msPrvTrigger();
	for (i = 0; i < 8; i++)
		ret = ret * 2 + msPrvBitRd();
	
	return ret;
}

enum MsReadResult {
	MsReadOK,
	MsReadCorrected,
	MsReadUncorrectable,
	MsReadError,
};

struct MsOobData {
	uint8_t overwriFlg;
	uint8_t manageFlag;
	uint16_t logicalAddr;	//BE on media
};

#define MS_MG_FLG_SYSTEM			0x04	//system-used page (not user data)




//TPCs
#define MS_RD_LDATA					0x02
#define MS_RD_SDATA					0x03
#define MS_RD_REG					0x04
#define MS_GET_INT					0x07
#define MS_SET_RW_REG_ADRS			0x08		//might not work on non-PRO MSs
#define MS_EX_SET_CMD				0x09		//might not work on non-PRO MSs
#define MS_WR_REG					0x0b
#define MS_WR_SDATA					0x0c
#define MS_WR_LDATA					0x0d
#define MS_SET_CMD					0x0e

//MS regs (for MS_RD_REG/MS_WR_REG)
#define MS_REG_INT					0x01	//RO
#define MS_REG_STA0					0x02	//RO
#define MS_REG_STA1					0x03	//RO
#define MS_REG_TYPE					0x04	//RO
#define MS_REG_IFACE				0x05	//RO
#define MS_REG_CATEGORY				0x06	//RO
#define MS_REG_CLASS				0x07	//RO

#define MS_REG_UNK_HI				0x09	//RO	io related. length of io attributes?
#define MS_REG_UNK_LO				0x0A	//RO	io related. length of io attributes?


#define MS_REG_SYSPARAM				0x10
#define MS_REG_BLK_ADR_HI			0x11
#define MS_REG_BLK_ADR_MID			0x12
#define MS_REG_BLK_ADR_LO			0x13
#define MS_REG_ACCESSTYP			0x14	//0x00 - block, 0x20 - page, 0x40 - OOB only, 0x80 - overwrite
#define MS_REG_PG_ADDR				0x15
#define MS_REG_OVERWRI_FLG			0x16	//struct MsOobData:
#define MS_REG_MANAGE_FLG			0x17
#define MS_REG_LOGICAL_ADR_HI		0x18
#define MS_REG_LOGICAL_ADR_LO		0x19
///5 more bytes here that are unused but are stored per block with ECC
#define MS_REG_READ_FIRST			MS_REG_INT
#define MS_REG_READ_LAST			MS_REG_LOGICAL_ADR_LO
#define MS_REG_WRITE_FIRST			MS_REG_SYSPARAM
#define MS_REG_WRITE_LAST			MS_REG_LOGICAL_ADR_LO
#define MS_NUM_REGS_READ			(MS_REG_READ_LAST + 1 - MS_REG_READ_FIRST)
#define MS_NUM_REGS_WRITE			(MS_REG_WRITE_LAST + 1 - MS_REG_WRITE_FIRST)

//sta1 bits
#define MS_STA1_UNCORRECTABLE		0x15
#define MS_STA1_CORRECTED			0x2a

//MSPRO regs (for MS_RD_REG)
#define MS_PRO_REG_INT				0x01	//RO
#define MS_PRO_REG_STATUS			0x02	//RO
#define MS_PRO_REG_TYPE				0x04	//RO
#define MS_PRO_REG_CATEGORY			0x06	//RO
#define MS_PRO_REG_CLASS			0x07	//RO

//MSPRO regs (for MS_WR_REG)
#define MS_PRO_REG_SYSPARAM			0x10	//WO
#define MS_PRO_REG_DAT_CNT_1		0x11	//WO		num 512-byte "units" to xfer. BE. if zero, read/write will not end till MS_PRO_CMD_STOP
#define MS_PRO_REG_DAT_CNT_0		0x12	//WO
#define MS_PRO_REG_DAT_ADR_3		0x13	//WO		address (in units?). BE
#define MS_PRO_REG_DAT_ADR_2		0x14	//WO
#define MS_PRO_REG_DAT_ADR_1		0x15	//WO
#define MS_PRO_REG_DAT_ADR_0		0x16	//WO
#define MS_PRO_REG_TPC_PARAM		0x17	//WO		short xfer size in lower 2 bits: {0->32, 1->64, 2->128, 3->256}
#define MS_PRO_REG_CMD_PARAM		0x18	//WO

//MS_SET_RW_REG_ADRS writes:
// READ_ADRS W Set the starting address accessed by a READ_REG TPC.
// READ_SIZE W Set the number of bytes accessed by a READ_REG TPC.
// WRITE_ ADRS W Set the starting address accessed by a WRITE_REG TPC.
// WRITE_ SIZE W Set the number of bytes accessed by a WRITE_REG TPC.
//basically read/write reg access a range. this is the range!
//MSPRO defaults to 00 08 10 10
//MS defaults to 00 1F 10 0F

#define MS_CMD_RESET				0x3c
#define MS_CMD_PAGE_READ			0xaa

#define MS_PRO_CMD_FORMAT			0x10
#define MS_PRO_CMD_READ				0x20
#define MS_PRO_CMD_WRITE			0x21
#define MS_PRO_CMD_ATTRIB_READ		0x24
#define MS_PRO_CMD_STOP				0x25

//CRC poly is x^16 + x^15 + x^2 + 1 -> crc-16 with poly 0x8005
//crc init value is 0		(CRC-16/BUYPASS seems to match)

static uint16_t msPrvCrc(uint8_t byte, uint16_t crc)
{
	static const uint16_t crcTab[] = {
		0x0000, 0x8005, 0x800F, 0x000A, 0x801B, 0x001E, 0x0014, 0x8011,
		0x8033, 0x0036, 0x003C, 0x8039, 0x0028, 0x802D, 0x8027, 0x0022,
		0x8063, 0x0066, 0x006C, 0x8069, 0x0078, 0x807D, 0x8077, 0x0072,
		0x0050, 0x8055, 0x805F, 0x005A, 0x804B, 0x004E, 0x0044, 0x8041,
		0x80C3, 0x00C6, 0x00CC, 0x80C9, 0x00D8, 0x80DD, 0x80D7, 0x00D2,
		0x00F0, 0x80F5, 0x80FF, 0x00FA, 0x80EB, 0x00EE, 0x00E4, 0x80E1,
		0x00A0, 0x80A5, 0x80AF, 0x00AA, 0x80BB, 0x00BE, 0x00B4, 0x80B1,
		0x8093, 0x0096, 0x009C, 0x8099, 0x0088, 0x808D, 0x8087, 0x0082,
		0x8183, 0x0186, 0x018C, 0x8189, 0x0198, 0x819D, 0x8197, 0x0192,
		0x01B0, 0x81B5, 0x81BF, 0x01BA, 0x81AB, 0x01AE, 0x01A4, 0x81A1,
		0x01E0, 0x81E5, 0x81EF, 0x01EA, 0x81FB, 0x01FE, 0x01F4, 0x81F1,
		0x81D3, 0x01D6, 0x01DC, 0x81D9, 0x01C8, 0x81CD, 0x81C7, 0x01C2,
		0x0140, 0x8145, 0x814F, 0x014A, 0x815B, 0x015E, 0x0154, 0x8151,
		0x8173, 0x0176, 0x017C, 0x8179, 0x0168, 0x816D, 0x8167, 0x0162,
		0x8123, 0x0126, 0x012C, 0x8129, 0x0138, 0x813D, 0x8137, 0x0132,
		0x0110, 0x8115, 0x811F, 0x011A, 0x810B, 0x010E, 0x0104, 0x8101, 
		0x8303, 0x0306, 0x030C, 0x8309, 0x0318, 0x831D, 0x8317, 0x0312,
		0x0330, 0x8335, 0x833F, 0x033A, 0x832B, 0x032E, 0x0324, 0x8321,
		0x0360, 0x8365, 0x836F, 0x036A, 0x837B, 0x037E, 0x0374, 0x8371,
		0x8353, 0x0356, 0x035C, 0x8359, 0x0348, 0x834D, 0x8347, 0x0342,
		0x03C0, 0x83C5, 0x83CF, 0x03CA, 0x83DB, 0x03DE, 0x03D4, 0x83D1,
		0x83F3, 0x03F6, 0x03FC, 0x83F9, 0x03E8, 0x83ED, 0x83E7, 0x03E2,
		0x83A3, 0x03A6, 0x03AC, 0x83A9, 0x03B8, 0x83BD, 0x83B7, 0x03B2,
		0x0390, 0x8395, 0x839F, 0x039A, 0x838B, 0x038E, 0x0384, 0x8381,
		0x0280, 0x8285, 0x828F, 0x028A, 0x829B, 0x029E, 0x0294, 0x8291,
		0x82B3, 0x02B6, 0x02BC, 0x82B9, 0x02A8, 0x82AD, 0x82A7, 0x02A2,
		0x82E3, 0x02E6, 0x02EC, 0x82E9, 0x02F8, 0x82FD, 0x82F7, 0x02F2,
		0x02D0, 0x82D5, 0x82DF, 0x02DA, 0x82CB, 0x02CE, 0x02C4, 0x82C1,
		0x8243, 0x0246, 0x024C, 0x8249, 0x0258, 0x825D, 0x8257, 0x0252,
		0x0270, 0x8275, 0x827F, 0x027A, 0x826B, 0x026E, 0x0264, 0x8261,
		0x0220, 0x8225, 0x822F, 0x022A, 0x823B, 0x023E, 0x0234, 0x8231,
		0x8213, 0x0216, 0x021C, 0x8219, 0x0208, 0x820D, 0x8207, 0x0202
	};

	return (crc << 8) ^ crcTab[(crc >> 8) ^ byte];
}

static bool msPrvTpc(uint_fast8_t tpc, void *data, uint_fast16_t len)
{
	uint8_t *ptr = (uint8_t*)data;
	uint_fast8_t v, i, tpcByte;
	uint_fast16_t crcCalc = 0;
	bool ret = false;
	
	pr("TPC 0x0%x with %u len\n", tpc, len);
	tpcByte = (tpc << 4) + ((~tpc) & 0x0f);
	
	msPrvBsToggle();	//go to BS1
	msPrvSdioOutput();
	msPrvBitWr(0, false);	//dummy
	
	//now send TPC and goto BS2
	msPrvByteWr(tpcByte, true);
	
	if (tpc & 0x8) {	//write TPC
		
		while (len--) {		//send data and go to BS3 if needed
			v = *ptr++;
			
			crcCalc = msPrvCrc(v, crcCalc);
			msPrvByteWr(v, false);
		}
		
		msPrvByteWr(crcCalc >> 8, false);
		msPrvByteWr((uint8_t)crcCalc, true);
		
		
		msPrvBitWr(0, false);	//dummy bits?
		msPrvSdioInput();
		msPrvBitWr(0, false);
		
		for (i = 0; i < 32; i++) {
			
			uint_fast8_t v = msPrvByteRd();
			
			if (v == 0xaa || v == 0x55) {
				
				ret = true;
				break;
			}
		}
	}
	else {				//read RPC
		
	//	msPrvBitWr(0, false);	//only one dummy bit needed here
		msPrvSdioInput();
		msPrvBitWr(0, false);
		
		for (i = 0; i < 64; i++) {
			
			uint_fast8_t v = msPrvByteRd();
			
			if (v == 0xaa || v == 0x55) {
				
				ret = true;
				break;
			}
		}
		
		msPrvBsToggle();	//go to BS3
		//dummy bit
		msPrvBitWr(0, false);
		
		while (len--) {
			
			v = msPrvByteRd();
			crcCalc = msPrvCrc(v, crcCalc);
			*ptr++ = v;
			
		}
		
		crcCalc ^= ((uint_fast16_t)msPrvByteRd()) << 8;
		crcCalc ^= msPrvByteRd();
		
		
		if (crcCalc)
			ret = false;
	}
	//we get here in read mode anyways
	msPrvBsToggle();	//go to BS0
	msPrvByteRd();
	
	return ret;
}

static bool msPrvCmdEx(uint_fast8_t cmd, uint_fast16_t count, uint32_t addr)	//waits for 0x20 INT reg, which may not be proper for MS_PRO_CMD_STOP
{
	uint8_t rx, tx[] = {cmd, count >> 8, (uint8_t)count, addr >> 24, addr >> 16, addr >> 8, (uint8_t)addr};
	
	if (!msPrvTpc(MS_EX_SET_CMD, tx, sizeof(tx))) {
		pr("cmd broken\n");
		return false;
	}
	
	do {
	
		if (!msPrvTpc(MS_GET_INT, &rx, 1)) {
			pr("get INT failed\n");
			return false;
		}
		pr("INT: %02x\n", rx);
		
	} while (rx != 0x20);	//wait for command complete
	
	return true;
}

static bool msPrvCmd(uint_fast8_t cmd, uint_fast8_t expectedIntVal)
{
	uint8_t rx, tx = cmd;
	
	if (!msPrvTpc(MS_SET_CMD, &tx, 1)) {
		pr("cmd broken\n");
		return false;
	}
	
	do {
	
		if (!msPrvTpc(MS_GET_INT, &rx, 1)) {
			pr("get INT failed\n");
			return false;
		}
		pr("INT: %02x\n", rx, cmd);
		
	} while (rx != expectedIntVal);	//wait for command complete
	
	return true;
}

static bool msPrvReadMsProAttrPage(uint_fast8_t pgIdx, uint8_t *buf)		//maybe MsPro only, buffer should be 514 bytes
{
	return msPrvCmdEx(MS_PRO_CMD_ATTRIB_READ, 1, pgIdx) && msPrvTpc(MS_RD_LDATA, buf, 512);
}

static bool msPrvReadAttr(uint32_t ofst, uint32_t len, uint8_t *dst)
{
	uint8_t buf[512];
	uint32_t now;
	
	while (len) {
		now = 512 - (ofst % 512);
		if (now > len)
			now = len;
		
		if (!msPrvReadMsProAttrPage(ofst / 512, buf))
			return false;
		memcpy(dst, buf + ofst % 512, now);
		dst += now;
		ofst += now;
		len -= now;
	}
	
	return true;
}

static uint_fast16_t msPrvReadBE16(const uint8_t *val)
{
	return (((uint_fast16_t)val[0]) << 8) + val[1];
}

static uint32_t msPrvReadBE32(const uint8_t *val)
{
	return (((uint32_t)val[0]) << 24) + (((uint32_t)val[1]) << 16) + (((uint32_t)val[2]) << 8) + val[3];
}

static void msPrvShowMsproAttr(uint_fast8_t id, const uint8_t *arr, uint32_t len)
{
	uint_fast8_t i;
	
	switch (id) {
		case 0x10:
			if (len != 96) {
				pr("invalid length\n");
				break;
			}
			pr("System Information block\n");
			pr(" Memory stick class:    %u\n", arr[0]);
			pr(" Block size:            %u\n", msPrvReadBE16(arr + 2));
			pr(" Total blocks:          %u\n", msPrvReadBE16(arr + 4));
			pr(" User area blocks:      %u\n", msPrvReadBE16(arr + 6));
			pr(" Page Size:             %u\n", msPrvReadBE16(arr + 8));
			pr(" Assembly date:         %04u.%02u.%02u %02u:%02u:%02u\n", msPrvReadBE16(arr + 13), arr[15], arr[16], arr[17], arr[18], arr[19]);
			pr(" Serial number:         %u\n", msPrvReadBE32(arr + 20));
			pr(" Assembly maker code:   %02xh\n", arr[24]);
			pr(" Assembly model code:   %02xh %02xh %02xh\n", arr[25], arr[26], arr[27]);
			pr(" Memory maker code:     %02xh %02xh\n", arr[28], arr[29]);
			pr(" Memory model code:     %02xh %02xh\n", arr[30], arr[31]);
			pr(" Vcc:                   %u\n", arr[36]);
			pr(" Vpp:                   %u\n", arr[37]);
			pr(" Controller number:     %02xh %02xh\n", arr[38], arr[39]);
			pr(" Controller function:   %02xh %02xh\n", arr[40], arr[41]);
			pr(" Start sector:          %u\n", msPrvReadBE16(arr + 42));
			pr(" Unit size:             %u\n", msPrvReadBE16(arr + 44));
			pr(" Memory stick subclass: %02xh\n", arr[46]);
			pr(" Interface type:        %02xh\n", arr[51]);
			pr(" Controller code:       %02xh %02xh\n", arr[52], arr[53]);
			pr(" Format type:           %02xh\n", arr[54]);
			pr(" Device type:           %02xh\n", arr[56]);
			pr(" Memory Stick PRO ID:  ");
			for (i = 0; i < 16; i++)
				pr(" %02x", arr[64 + i]);
			pr("\n");
			pr(" TOTAL SIZE:            %u sectors (%llu bytes)\n",
				(uint32_t)msPrvReadBE16(arr + 2) * msPrvReadBE16(arr + 6),
				((uint64_t)((uint32_t)msPrvReadBE16(arr + 2) * msPrvReadBE16(arr + 6))) * 512);
			break;
		
		case 0x15:
			pr(" MODEL: '%s'\n", arr);
			break;
		
		case 0x20:
			pr("PARTENTRY\n");
			break;
		
		case 0x21:
		case 0x22:
			pr("FAT%uHDR\n", 16 << (id - 0x21));
			break;
		
		case 0x25:
		case 0x26:
			pr("MS file values %u\n", id - 0x24);
			break;
	
		default:
			pr("unknown attribute 0x%02x\n", id);
			break;
	}	
}

static enum MsReadResult msPrvReadPage(uint32_t block, uint8_t page, uint8_t *data, struct MsOobData *oob)
{
	union {	//to save stack
		uint8_t wrRegs[MS_NUM_REGS_WRITE];
		uint8_t rdRegs[MS_NUM_REGS_READ];
		
	} regs = {
			.wrRegs = {
				[MS_REG_BLK_ADR_HI - MS_REG_WRITE_FIRST] = block >> 16,
				[MS_REG_BLK_ADR_MID - MS_REG_WRITE_FIRST] = block >> 8,
				[MS_REG_BLK_ADR_LO - MS_REG_WRITE_FIRST] = block >> 0,
				[MS_REG_ACCESSTYP - MS_REG_WRITE_FIRST] = data ? 0x20 : 0x40,
				[MS_REG_PG_ADDR - MS_REG_WRITE_FIRST] = page,
			},
		};
	
	//set regs for the read
	
	if (!msPrvTpc(MS_WR_REG, regs.wrRegs, MS_NUM_REGS_WRITE)) {
		pr("failed to write regs\n");
		return MsReadError;
	}
	
	if (!msPrvCmd(MS_CMD_PAGE_READ, 0xa0)) {
		pr("failed to initiate the command\n");
		return MsReadError;
	}
	
	//we need the status regs in all cases to see correctability info, so always read the regs
	if (!msPrvTpc(MS_RD_REG, regs.rdRegs, MS_NUM_REGS_READ)) {
		pr("get REGS fail\n");
		return MsReadError;
	}
	pr("REGS:\n");
	showBuf(regs.rdRegs, MS_NUM_REGS_READ);
	
	//if OOB was requested, provide it
	if (oob) {
		
		oob->overwriFlg = regs.rdRegs[MS_REG_OVERWRI_FLG - MS_REG_READ_FIRST];
		oob->manageFlag = regs.rdRegs[MS_REG_MANAGE_FLG - MS_REG_READ_FIRST];
		oob->logicalAddr = msPrvReadBE16(regs.rdRegs + MS_REG_LOGICAL_ADR_HI - MS_REG_READ_FIRST);
	}
	
	//if data was requested, read that
	if (data) {
		
		if (!msPrvTpc(MS_RD_LDATA, data, 512)) {
			pr("get FATA fail\n");
			return MsReadError;
		}
	}
	
	//sort out correctability
	if (regs.rdRegs[MS_REG_STA1 - MS_REG_READ_FIRST] & MS_STA1_UNCORRECTABLE)
		return MsReadUncorrectable;
	
	if (regs.rdRegs[MS_REG_STA1 - MS_REG_READ_FIRST] & MS_STA1_CORRECTED)
		return MsReadCorrected;
	
	return MsReadOK;
}

static void msTest(void)
{
	static uint8_t buf[512];
	struct MsOobData oob;
	enum MsReadResult rr;
	uint_fast8_t block;
	
	buf[0] = MS_REG_READ_FIRST;
	buf[1] = MS_NUM_REGS_READ;
	buf[2] = MS_REG_WRITE_FIRST;
	buf[3] = MS_NUM_REGS_WRITE;
	
	if (!msPrvTpc(MS_SET_RW_REG_ADRS, buf, 4)) {
		pr("failed to set reg addrs\n");
		return;
	}
	
	if (!msPrvCmd(MS_CMD_RESET, 0x00)) {
		pr("failed to initiate the reset command\n");
		return;
	}
	
	//read info page. not so easy for MS since we have to do software wear leveling, and we do not even know where to look
	//we look in the first 17 blocks, same as the Clie driver does
	for (block = 0; block < 17; block++) {
		
		rr = msPrvReadPage(block, 0, buf, &oob);
		if (rr == MsReadError) {
			pr("read page error\n");
			return;
		}
		
		pr("read %u.0: ret = %u, OOB = {ov=%02xh, mng=%02xh, la=%04xh}\n", block, rr, oob.overwriFlg, oob.manageFlag, oob.logicalAddr);
		showBuf(buf, 512);

		if (rr == MsReadUncorrectable) {
			pr("%u.0 is uncorrectable\n", block);
			continue;
		}
		
		if (oob.manageFlag & MS_MG_FLG_SYSTEM) {
			pr("%u.0 has unexpected flags for info page\n", block);
			continue;
		}
		
		if (buf[0] != 0x00 || buf[1] != 0x01) {
			pr("%u.0 has unexpected first bytes for info page\n", block);
			continue;
		}
		
		pr("possible boot block flagged\n");
	}
}



static void msproTest(void)
{
	static uint8_t vals[512];
	uint_fast8_t numAttrs, i;
	
	vals[0] = 0x00;
	vals[1] = 0x08;
	vals[2] = 0x10;
	vals[3] = 0x10;
	
	if (!msPrvTpc(MS_SET_RW_REG_ADRS, vals, 4)) {
		pr("failed to set reg addrs\n");
		return;
	}
	

	do {
		msPrvTpc(MS_RD_REG, vals, 8);
	} while (!(vals[1] & 0x80));
	
	pr("REGS:");
	showBuf(vals, 8);
	pr("card inited\n");
	
	if (!msPrvReadMsProAttrPage(0, vals)) {
		pr("failed to read attrs page\n");
		return;
	}
	showBuf(vals, 512);
	
	//parse attribs
	if (vals[0] != 0xa5 || vals[1] != 0xc3 || vals[2] != 0x01 || (numAttrs = vals[4]) > 12) {
		pr("attrib page unparseable\n");
		return;
	}
	for (i = 0; i < numAttrs; i++) {
		uint32_t ofst = msPrvReadBE32(vals + 16 + 12 * i + 0);
		uint32_t len = msPrvReadBE32(vals + 16 + 12 * i + 4);
		uint_fast8_t id = vals[16 + 12 * i + 8];
		uint8_t arr[96];	//big enough?
		
		pr("getting attr %u (id %u) (0x%08x + 0x%08x)\n", i, id, ofst, len);
		
		if (ofst < 0x1a0 || ofst + len > 0x8000) {
			pr("invalid attr\n");
			return;
		}
		
		if (len > sizeof(arr)) {
			
			pr("too big\n");
			continue;
		}
		
		if (!msPrvReadAttr(ofst, len, arr)) {
			
			pr("failed to read attribute\n");
			return;
		}
		
		showBuf(arr, len);
		msPrvShowMsproAttr(id, arr, len);
	}
	
	//let's try to read some sectors?
	if (!msPrvCmdEx(MS_PRO_CMD_READ, 2, 0)) {
		pr("failed to set up read command\n");
		return;
	}
	
	//let's read data
	if (!msPrvTpc(MS_RD_LDATA, vals, 512)) {
		pr("failed to set up read pg 0\n");
		return;
	}
	showBuf(vals, 512);
	
	if (!msPrvTpc(MS_RD_LDATA, vals, 512)) {
		pr("failed to set up read pg 1\n");
		return;
	}
	showBuf(vals, 512);
	
	//let's write data
	vals[5] = 0x66;
	vals[6] = 0x77;
	vals[7] = 0x88;
	
	//let's try to read some sectors?
	if (!msPrvCmdEx(MS_PRO_CMD_WRITE, 1, 1)) {
		pr("failed to set up write command\n");
		return;
	}
	
	if (!msPrvTpc(MS_WR_LDATA, vals, 512)) {
		pr("failed to set up wr pg 1\n");
		return;
	}
	
	pr("wrote\n");
	memset(vals, 0, sizeof(vals));
	
	pr("readback\n");
	if (!msPrvCmdEx(MS_PRO_CMD_READ, 1, 1)) {
		pr("failed to set up read command\n");
		return;
	}
	
	if (!msPrvTpc(MS_RD_LDATA, vals, 512)) {
		pr("failed to set up read pg 1\n");
		return;
	}
	showBuf(vals, 512);
}

static void msAnyTest(void)
{
	static uint8_t vals[4];
	
	vals[0] = MS_REG_TYPE;
	vals[1] = 1;
	vals[2] = 0x10;
	vals[3] = 0x10;
	
	if (!msPrvTpc(MS_SET_RW_REG_ADRS, vals, 4)) {
		pr("failed to set reg addrs\n");
		return;
	}
	
	msPrvTpc(MS_RD_REG, vals, 1);
	
	pr("TYPE: %02xh\n", vals[0]);
	
	if (vals[0] == 0xff)
		msTest();
	else if (vals[0] == 0x00)
		pr("what do you expect me to do with an MSIO stick?\n");
	else if (vals[0] == 0x01)
		msproTest();
	else
		pr("not sure what to do about this device type\n");
}

int main(void)
{
	volatile uint32_t t = 0;
	
	pr("hello\n");
	
	initHw();
	timebaseInit();
	pr("ready, time is 0x%016llx\n", getTime());
	
	msAnyTest();
	
	while(1);
}
